use crate::{ruleset::{Card, CardMetadata, Setup}, dumb_dealer::Deal, zobrist::{Feature, Zobrist}};

#[derive(Clone, Copy, Debug, Hash)]
pub enum Move {
    ToSlot { from_slot: u8, to_slot: u8 },
    ToWell { from_slot: u8, to_well: u8 }
}


pub struct Board<'a> {
    zobrist: Zobrist,

    checkpoints: Vec<usize>,
    log: Vec<Move>,

    setup: &'a Setup,
    slots: Vec<Slot<'a>>,
    wells: Vec<Well<'a>>,
}

impl<'a> Board<'a> {
    pub fn new(setup: &'a Setup) -> Self {
        let mut slots = vec![];
        let mut wells = vec![];

        // ..n_slots: normal
        for _ in 0..setup.ruleset.n_slots {
            slots.push(Slot::new(setup, slots.len() as u8, 32));
        }
        // n_slots: top cell
        slots.push(Slot::new(setup, slots.len() as u8, 1));

        // ..n_suits: conventional wells
        for _ in &setup.deck.suits {
            wells.push(Well::new(setup, wells.len() as u8, Box::new(
                |prev: Option<CardMetadata>, new: CardMetadata| {
                    if let Some(p) = prev {
                        if new.suit != p.suit { return false }
                        return new.suit == p.suit && new.rank == p.rank + 1
                    }
                    panic!("well should have had an ace or something to start")
                }
            )))
        }

        if setup.ruleset.n_arcana > 0 {
            // n_suits: arcana ascending
            wells.push(Well::new(setup, wells.len() as u8, Box::new(
                |prev, new| {
                    if new.suit != b'a' { return false }
                    if let Some(p) = prev {
                        return new.rank == p.rank + 1;
                    }
                    return new.rank == 0;
                }
            )));

            // n_suits+1: arcana descending
            let top_rank = setup.ruleset.n_arcana - 1;
            wells.push(Well::new(setup, wells.len() as u8, Box::new(
                move |prev, new| {
                    if new.suit != b'a' { return false }
                    if let Some(p) = prev {
                        return new.rank + 1 == p.rank;
                    }
                    return new.rank == top_rank;
                }
            )))
        }
        Board {
            zobrist: Zobrist::new(),

            checkpoints: vec![],
            log: vec![],

            setup,
            slots,
            wells,
        }
    }

    pub fn display(&self) {
        fn display_cards(setup: &Setup, cards: &[Card]) {
            print!("- ");
            for c in cards {
                let meta = setup.deck.cards[c.0 as usize];
                print!("{}{:02} ", meta.suit as char, meta.rank);
            }
            print!("\n");
        }
        println!("Wells:");
        for w in &self.wells {
            display_cards(self.setup, &w.contents);
        }
        println!("\nSlots:");
        for s in &self.slots {
            display_cards(self.setup, &s.contents);
        }
    }

    pub fn zobrist_key(&self) -> u64 {
        self.zobrist.value
    }

    pub fn is_won(&self) -> bool {
        for s in self.slots.iter() {
            if s.peek().is_some() {
                return false
            }
        }
        return true
    }

    pub fn deal(&mut self, deal: Deal) {
        let n_usable_slots = self.setup.ruleset.n_slots - 1;

        for (w, c) in self.setup.deck.aces.iter().enumerate() {
            self.wells[w].push(&mut self.zobrist, *c);
        }

        for (s, content) in deal.slots.iter().enumerate() {
            let real_slot = if s < n_usable_slots as usize / 2 { s } else { s + 1 };
            for &card in content {
                self.slots[real_slot].push(&mut self.zobrist, card)
            }
        }
    }

    fn settle(&mut self) {
        while self.settle1() { }
    }

    fn settle1(&mut self) -> bool {
        let special_slot_full = self.slots[self.setup.ruleset.n_slots as usize].peek().is_some();
        for s in 0..self.slots.len() {
            if let Some(top) = self.slots[s].peek() {
                for w in 0..self.wells.len() {
                    if special_slot_full && (w as u8) < self.setup.ruleset.n_suits {
                        // top wells are blocked
                    } else if self.wells[w].would_accept(top) {
                        self.internal_perform(Move::ToWell { from_slot: s as u8, to_well: w as u8 });
                        return true
                    }
                }
            }
        }
        return false;
    }

    // TODO: Do this without allocs?
    pub fn legal_moves(&self) -> Vec<Move> {
        let mut moves = vec![];
        for i in 0..self.slots.len() { 
            if let Some(top) = self.slots[i].peek() {
                for j in 0..self.slots.len() {
                    if i == j { continue; }
                    if !self.slots[j].would_accept(top) { continue; }
                    moves.push(Move::ToSlot { from_slot: i as u8, to_slot: j as u8 })
                }
            }
        }
        moves
    }

    fn create_checkpoint(&mut self) {
        self.checkpoints.push(self.log.len());
    }

    pub fn perform(&mut self, m: Move) {
        self.create_checkpoint();
        self.internal_perform(m);
        self.settle();
    }

    pub fn undo(&mut self) {
        let cp = self.checkpoints.pop().expect("undo() without perform()");
        while self.log.len() > cp {
            let m = self.log.pop().unwrap();
            self.internal_unperform(m)
        }
    }

    fn internal_perform(&mut self, m: Move) {
        // println!("internal perform: {:?}", m);
        match m {
            Move::ToWell { from_slot, to_well } => {
                let top = self.slots[from_slot as usize].pop(&mut self.zobrist).expect("move from empty slot");
                self.wells[to_well as usize].push(&mut self.zobrist, top);
            }
            Move::ToSlot { from_slot, to_slot } => {
                let top = self.slots[from_slot as usize].pop(&mut self.zobrist).expect("move from empty slot");
                self.slots[to_slot as usize].push(&mut self.zobrist, top);
            }
        }
        self.log.push(m);
    }

    fn internal_unperform(&mut self, m: Move) {
        // println!("internal unperform: {:?}", m);
        match m {
            Move::ToWell { from_slot, to_well } => {
                let top = self.wells[to_well as usize].pop(&mut self.zobrist).expect("move from empty well");
                self.slots[from_slot as usize].push(&mut self.zobrist, top);
            }
            Move::ToSlot { from_slot, to_slot } => {
                let top = self.slots[to_slot as usize].pop(&mut self.zobrist).expect("move from empty slot");
                self.slots[from_slot as usize].push(&mut self.zobrist, top);
            }
        }
    }
}

struct Slot<'a> {
    ix: u8,
    setup: &'a Setup,
    max_n: u8,
    contents: Vec<Card>,
}

impl<'a> Slot<'a> {
    fn new(setup: &'a Setup, ix: u8, max_n: u8) -> Self {
        Slot {
            setup,
            ix,
            contents: Vec::with_capacity(max_n as usize),
            max_n,
        }
    }

    fn push(&mut self, zobrist: &mut Zobrist, new: Card) {
        zobrist.toggle(Feature::CardAt { card: new, slot: self.ix, depth: self.contents.len() as u8});
        self.contents.push(new);
    }

    fn peek(&self) -> Option<Card> {
        self.contents.last().cloned()
    }

    fn pop(&mut self, zobrist: &mut Zobrist) -> Option<Card> {
        let card = self.contents.pop();
        let depth = self.contents.len() as u8;
        if let Some(card) = card {
            zobrist.toggle(Feature::CardAt { card, slot: self.ix, depth });
        }
        return card
    }

    fn would_accept(&self, top: Card) -> bool {
        if self.contents.len() as u8 >= self.max_n {
            return false;
        }

        let last = self.contents.last();
        if let Some(l) = last {
            let l = self.setup.deck.cards[l.0 as usize];
            let n = self.setup.deck.cards[top.0 as usize];
            return l.suit == n.suit && (l.rank + 1 == n.rank || l.rank == n.rank + 1);
        } else {
            return true;
        }
    }
}
type AcceptCb = Box<dyn Fn(Option<CardMetadata>, CardMetadata) -> bool>;

struct Well<'a> {
    setup: &'a Setup,
    ix: u8,
    accept_cb: AcceptCb,
    contents: Vec<Card>,
}

impl<'a> Well<'a> {
    fn new(setup: &'a Setup, ix: u8, accept_cb: AcceptCb) -> Self {
        Well {
            setup,
            ix,
            accept_cb,
            contents: vec![]
        }
    }

    fn push(&mut self, zobrist: &mut Zobrist, new: Card) {
        zobrist.toggle(Feature::WellHas { card: new, well: self.ix });
        self.contents.push(new);
    }

    fn pop(&mut self, zobrist: &mut Zobrist) -> Option<Card> {
        let card = self.contents.pop();
        if let Some(c) = card {
            zobrist.toggle(Feature::WellHas { card: c, well: self.ix });
        }
        card
    }

    fn would_accept(&self, new: Card) -> bool {
        let prev = self.contents.last().map(|p|
            self.setup.deck.cards[p.0 as usize]
        );
        let new = self.setup.deck.cards[new.0 as usize];
        return (self.accept_cb)(prev, new);
    }
}